import math
import struct
import zlib
import time
import sys
sys.path
import Conf
sys.path.append("misc")
from other import *


#######################################################################
# Related to the composing and checking of packets

class JobPacket:

    def __init__(self):
        """
        Defines default values, related to the generation of packets
        """
        self.commands = ""
        self.fullsps = 0    # number of subpackets - 1
        self.lastspsize = 0    # number of commands on last subpacket
        self.width = 0      # calculated width for perimeter
        self.height = 0     # calculated height for perimeter
        self.header = []    # where will be the main header
        self.subheader = ['\x00'] * 8    # where will be the header for cut commands
        self.packetdata = ""    # stores important info about the packet
        self.ok = True      # True when there are no errors in packet
        self.totalcmds = 0      # amount of found commands
        self.datacmds = ""      # info about found commands
        self.decompressed = ""      # result from decompressing subpackets, subheader + commands
        self.widthheader = 0    # width according to header
        self.heightheader = 0   # height according to header
        self.powerheader = 0    # power according to header
        self.numberstructsheader = 0    # number of structs according to subheader
        self.binaryjob = ""     # stores the binary to save/send or the binary to check
        self.foundcmds = None       # stores the type of present commands and their respective ocurrencies
        self.cantsubpack = 0    # given by packet

    ####################################################################
    # Methods related to the composing of the packet

    def makepacket(self, cbytes):
        """
        Accepts the commands of the cut as a string and
        calls all the methods necessary to make the full packet
        if exceeds the size then returns None
        :param cbytes: binary stream with just the commands of the job
        :rtype: str
        :return: the binary job, ready to be written to a file or sent to the machine
        """

        self.commands = cbytes  # binary stream
        self.fullsps = len(self.commands) / 0x40000     # calculates number of subpackets - 1
        self.lastspsize = (len(self.commands) % 0x40000) / 4     # calculates number of commands on last subpacket

        # checks the size of the commands
        if len(self.commands) > 0x1800000:
            print ("Job out of memory")
            return

        # sets the subheader
        self.subheader[0] = '\x02'
        self.subheader[4:6] = struct.pack('<H', self.lastspsize)
        self.subheader[6] = struct.pack('B', self.fullsps)

        # joins the subheader and the commands
        self.decompressed = arr2str(self.subheader) + self.commands

        # calculate width and height from the current commands
        self.countcommands()
        self.calculatewh()

        # elaborate the principal header
        self.makeheader()

        # appends the header + number of subpackets + 00 00 00 , to the final binary
        self.binaryjob = self.getheader() + struct.pack('B', self.fullsps + 1) + '\x00\x00\x00'

        # appends the compressed strings to the final binary
        self.binaryjob += self.makecompressedsubs()

        return self.binaryjob

    def makeheader(self):
        """
        Makes the principal header of the job packet
        """
        # where would be the header
        self.header = ['\x00'] * 1024

        # value defined for a vector cut
        self.header[0] = '\x02'

        # value defined independent from parameters
        self.header[512] = '\x05'

        # adds configuration structure
        self.header[532:668] = self.makeconfstruct(self.width, self.height)

        # value defined independent from parameters
        self.header[668] = '\x17'

    def getheader(self):
        """
        Returns header as a string
        """
        return arr2str(self.header)

    @staticmethod
    def makeconfstruct(width=0, height=0):
        """
        Generates the config struct from Conf file and from
        perimeter given by width and height in steps and it returns
        the config struct as a list
        :rtype: list
        :param width: specifies the with of the cut on steps
        :param height: specifies the height of the cut on steps
        :return: configuration struct as a list of hex strings
        """
        configparam = ['\x00'] * 136

        if Conf.RotaryCheck:
            configparam[4:8] = struct.pack('<I', g(Conf.RotaryEngraveAcceleration[0]))
            configparam[8:12] = struct.pack('<I', g(Conf.RotaryEngraveAcceleration[1]))
            configparam[12:16] = struct.pack('<I', g(Conf.RotaryEngraveAcceleration[2]))

            configparam[16:20] = struct.pack('<I', Conf.RotaryEngraveMaxVelocity[0])
            configparam[20:24] = struct.pack('<I', Conf.RotaryEngraveMaxVelocity[1])
            configparam[24:28] = struct.pack('<I', Conf.RotaryEngraveMaxVelocity[2])

            configparam[28:32] = struct.pack('<I', g(Conf.RotaryVectorAcceleration[0]))
            configparam[32:36] = struct.pack('<I', g(Conf.RotaryVectorAcceleration[1]))
            configparam[36:40] = struct.pack('<I', g(Conf.RotaryVectorAcceleration[2]))

            configparam[40:44] = struct.pack('<I', Conf.RotaryVectorMaxVelocity[0])
            configparam[44:48] = struct.pack('<I', Conf.RotaryVectorMaxVelocity[1])
            configparam[48:52] = struct.pack('<I', Conf.RotaryVectorMaxVelocity[2])

            configparam[52:56] = struct.pack('<I', g(Conf.RotaryJogAcceleration[0]))
            configparam[56:60] = struct.pack('<I', g(Conf.RotaryJogAcceleration[1]))
            configparam[60:64] = struct.pack('<I', g(Conf.RotaryJogAcceleration[2]))

            configparam[64:68] = struct.pack('<I', Conf.RotaryJogMaxVelocity[0])
            configparam[68:72] = struct.pack('<I', Conf.RotaryJogMaxVelocity[1])
            configparam[72:76] = struct.pack('<I', Conf.RotaryJogMaxVelocity[2])

        else:
            configparam[4:8] = struct.pack('<I', g(Conf.EngraveAcceleration[0]))
            configparam[8:12] = struct.pack('<I', g(Conf.EngraveAcceleration[1]))
            configparam[12:16] = struct.pack('<I', g(Conf.EngraveAcceleration[2]))

            configparam[16:20] = struct.pack('<I', Conf.EngraveMaxVelocity[0])
            configparam[20:24] = struct.pack('<I', Conf.EngraveMaxVelocity[1])
            configparam[24:28] = struct.pack('<I', Conf.EngraveMaxVelocity[2])

            configparam[28:32] = struct.pack('<I', g(Conf.VectorAcceleration[0]))
            configparam[32:36] = struct.pack('<I', g(Conf.VectorAcceleration[1]))
            configparam[36:40] = struct.pack('<I', g(Conf.VectorAcceleration[2]))

            configparam[40:44] = struct.pack('<I', Conf.VectorMaxVelocity[0])
            configparam[44:48] = struct.pack('<I', Conf.VectorMaxVelocity[1])
            configparam[48:52] = struct.pack('<I', Conf.VectorMaxVelocity[2])

            configparam[52:56] = struct.pack('<I', g(Conf.JogAcceleration[0]))
            configparam[56:60] = struct.pack('<I', g(Conf.JogAcceleration[1]))
            configparam[60:64] = struct.pack('<I', g(Conf.JogAcceleration[2]))

            configparam[64:68] = struct.pack('<I', Conf.JogMaxVelocity[0])
            configparam[68:72] = struct.pack('<I', Conf.JogMaxVelocity[1])
            configparam[72:76] = struct.pack('<I', Conf.JogMaxVelocity[2])

        configparam[80:84] = struct.pack('<I', width)
        configparam[84:88] = struct.pack('<I', height)

        configparam[88] = '\x02'
        configparam[92] = '\x02'
        configparam[96] = '\x01'

        configparam[100:105] = [struct.pack('B', Conf.byte0),
                                struct.pack('B', Conf.FlipLaserOutput),
                                struct.pack('B', Conf.FlipLaserPWMPower),
                                struct.pack('B', Conf.LaserPowerMax),
                                struct.pack('B', Conf.HomeDirection)
                                ]

        configparam[105:108] = [struct.pack('B', Conf.FlipHomeDirection[0]),
                                struct.pack('B', Conf.FlipHomeDirection[1]),
                                struct.pack('B', Conf.FlipHomeDirection[2])
                                ]

        configparam[108:112] = [struct.pack('B', Conf.LimitContCondition[0]),
                                struct.pack('B', Conf.LimitContCondition[1]),
                                struct.pack('B', Conf.LimitContCondition[2]),
                                struct.pack('B', Conf.LimitContCondition[3]),
                                ]
        configparam[112:116] = struct.pack('<I', Conf.MaxSteps[0])
        configparam[116:120] = struct.pack('<I', Conf.MaxSteps[1])
        configparam[120:124] = struct.pack('<I', Conf.MaxSteps[2])

        configparam[124:128] = struct.pack('<I', Conf.TableSize[0])
        configparam[128:132] = struct.pack('<I', Conf.TableSize[1])
        configparam[132:136] = struct.pack('<I', Conf.TableSize[2])

        return configparam

    def makecompressedsubs(self):
        """
        Makes the concatenated compressed strings and returns them as a string
        :return: compressed strings ready to be added to packet
        :rtype: str
        """
        subps = []      # stores the compressed subpackets(strings)
        compressedjob = ""       # stores the resulting compressed strings in the form: size 00 00 compressedstring

        # adds the first packets except the last one
        for n in range(self.fullsps):
            subps += [self.getcompressed(self.decompressed[n * 0x40000:(n + 1) * 0x40000])]

        # adds the last one
        subps += [self.getcompressed(
            self.decompressed[self.fullsps * 0x40000:(self.fullsps + 1) * 0x40000 + self.lastspsize * 4])]

        # forms the pattern len(zstring0)+ 00 00 + zstring0 + len(zstring1)+ 00 00 + zstring1 ...
        for sub in subps:
            compressedjob += struct.pack('<H', len(sub)) + '\x00\x00' + sub

        return compressedjob

    @staticmethod
    def getcompressed(partjob):
        """
        Accepts a string and returns it compressed with zlib default and
        with the first two bytes trimmed.
        :param partjob: the subpacket to compress, must be a string
        :return: the result of compressing partjob, without the first two bytes
        :rtype: str
        """
        bytesz = zlib.compress(partjob)
        bytesz = bytesz[2:]
        return bytesz

    ####################################################################
    # Functions related to checking and parsing of existing packets

    def countcommands(self):
        """
        Receives the decompressed subpackets and founds the present commands and the number of times they occur,
        the generated info is included on details output, you may also access info about the commands and
        the times they occur with the dictionary self.foundcmds, must run readcheckpacket() or checkpacket() previously
        """

        decompressed = self.decompressed[:]

        # checks the subheader
        if decompressed[0:4] != '\x02\x00\x00\x00':
            self.ok = False
            self.packetdata += "\nAnormal bytes on " + str([0, 3]) + " positions :" + str([decompressed[0:4]])
        if decompressed[7] != '\x00':
            self.ok = False
            self.packetdata += "\nAnormal byte on 7 position"

        # extracts the number of commands from subheader
        self.lastspsize = struct.unpack('<H', decompressed[4:6])[0]  # commands on last packet
        self.fullsps = struct.unpack('B', decompressed[6])[0]    # amount of full subpackets
        self.numberstructsheader = self.lastspsize + (0x10000 * self.fullsps)

        # trims the subheader
        decompressed = decompressed[8:]

        # extracts the existent commands
        presentcmds = []
        for n in range(len(decompressed) / 4):
            presentcmds += [decompressed[4 * n:4 * n + 4]]

        # here will be only one command of each type
        repeatedstructs = []
        repeatedstructs += [presentcmds[0]]

        # here will be the number of types that the corresponding command appear
        numberoftimes = [1]

        # find the repeated commands and increment their counter
        for n in range(1, len(presentcmds)):
            repeated = False
            for k in range(len(repeatedstructs)):
                if presentcmds[n] == repeatedstructs[k]:
                    numberoftimes[k] += 1
                    repeated = True
                    break
            if not repeated:
                repeatedstructs += [presentcmds[n]]
                numberoftimes += [1]

        # begins the info about commands
        self.datacmds = "Present commands: "

        # total of encountered commands
        self.totalcmds = 0

        # determines which are cutting commands
        self.datacmds += "\nCut commands: "
        for i in range(len(numberoftimes)):
            self.totalcmds += numberoftimes[i]
            if repeatedstructs[i][-1:] != '\x00':
                self.datacmds += "\nStruct: " + str([repeatedstructs[i]]) + ", times :" + str(numberoftimes[i])

        # determines non cut commands
        self.datacmds += "\nNon cut commands:"
        for j in range(len(numberoftimes)):
            if repeatedstructs[j][-1:] == '\x00':
                self.datacmds += "\nStruct: " + str([repeatedstructs[j]]) + ", times :" + str(numberoftimes[j])

        # saved as a dict for further use in calculatewh()
        self.foundcmds = dict(zip(repeatedstructs, numberoftimes))

        # Next values must be the same
        self.datacmds += "\nTotal structs " + str(self.totalcmds) +\
                         "\nAccording to header: " + str(self.numberstructsheader)

        # checks the concordance between the number of commands
        # and the value given on subheader
        if self.totalcmds != self.numberstructsheader:
            self.packetdata += "\nThe number of counted commands doesnt \
                    correspond with the number given on subheader"
            self.ok = False

    @staticmethod
    def parseconf(confstruct):
        """
        Reads a config structure and gets the relevant fields, made for testing porpuses
        :param confstruct: configuration struct as a binary string
        :return: info about the contents of the configuration struct
        :rtype: str
        """

        confdata = "Config struct: "
        engraveacc0 = ginv(struct.unpack('<I', confstruct[4:8])[0])
        confdata += "\nEngraveAcceleration[0]:"+str(engraveacc0)
        engraveacc1 = ginv(struct.unpack('<I', confstruct[8:12])[0])
        confdata += "\nEngraveAcceleration[1]:"+str(engraveacc1)
        engraveacc2 = ginv(struct.unpack('<I', confstruct[12:16])[0])
        confdata += "\nEngraveAcceleration[2]:"+str(engraveacc2)

        engravemaxvel0 = struct.unpack('<I', confstruct[16:20])[0]
        confdata += "\nEngraveMaxVelocity[0]:"+str(engravemaxvel0)
        engravemaxvel1 = struct.unpack('<I', confstruct[20:24])[0]
        confdata += "\nEngraveMaxVelocity[1]:"+str(engravemaxvel1)
        engravemaxvel2 = struct.unpack('<I', confstruct[24:28])[0]
        confdata += "\nEngraveMaxVelocity[2]:"+str(engravemaxvel2)

        vectoracc0 = ginv(struct.unpack('<I', confstruct[28:32])[0])
        confdata += "\nVectorAcceleration[0]:"+str(vectoracc0)
        vectoracc1 = ginv(struct.unpack('<I', confstruct[32:36])[0])
        confdata += "\nVectorAcceleration[1]:"+str(vectoracc1)
        vectoracc2 = ginv(struct.unpack('<I', confstruct[36:40])[0])
        confdata += "\nVectorAcceleration[2]:"+str(vectoracc2)

        vectormaxvel0 = struct.unpack('<I', confstruct[40:44])[0]
        confdata += "\nVectorMaxVelocity[0]:"+str(vectormaxvel0)
        vectormaxvel1 = struct.unpack('<I', confstruct[44:48])[0]
        confdata += "\nVectorMaxVelocity[1]:"+str(vectormaxvel1)
        vectormaxvel2 = struct.unpack('<I', confstruct[48:52])[0]
        confdata += "\nVectorMaxVelocity[2]:"+str(vectormaxvel2)

        jogacc0 = ginv(struct.unpack('<I', confstruct[52:56])[0])
        confdata += "\nJogAcceleration[0]:"+str(jogacc0)
        jogacc1 = ginv(struct.unpack('<I', confstruct[56:60])[0])
        confdata += "\nJogAcceleration[1]:"+str(jogacc1)
        jogacc2 = ginv(struct.unpack('<I', confstruct[60:64])[0])
        confdata += "\nJogAcceleration[2]:"+str(jogacc2)

        jogmaxvel0 = struct.unpack('<I', confstruct[64:68])[0]
        confdata += "\nJogMaxVelocity[0]:" + str(jogmaxvel0)
        jogmaxvel1 = struct.unpack('<I', confstruct[68:72])[0]
        confdata += "\nJogMaxVelocity[1]:" + str(jogmaxvel1)
        jogmaxvel2 = struct.unpack('<I', confstruct[72:76])[0]
        confdata += "\nJogMaxVelocity[2]:" + str(jogmaxvel2)

        width = struct.unpack('<I', confstruct[80:84])[0]
        height = struct.unpack('<I', confstruct[84:88])[0]
        confdata += "\nWidth:"+str(width)
        confdata += "\nHeight:"+str(height)

        byte0 = struct.unpack('B', confstruct[100])[0]
        fliplaserout = struct.unpack('B', confstruct[101])[0]
        fliplaserpwm = struct.unpack('B', confstruct[102])[0]
        laserpower = struct.unpack('B', confstruct[103])[0]
        homedirection = struct.unpack('B', confstruct[104])[0]
        confdata += "\nbyte0 FlipLaserOutput FlipLaserPWM LaserPowerMax HomeDirection:" + \
            str(byte0) + str(fliplaserout) + str(fliplaserpwm) + str(laserpower) + str(homedirection)

        fliphomedirection0 = struct.unpack('B', confstruct[105])[0]
        fliphomedirection1 = struct.unpack('B', confstruct[106])[0]
        fliphomedirection2 = struct.unpack('B', confstruct[107])[0]
        confdata += "\nFlipHomeDirection:" + str(fliphomedirection0) + \
                    str(fliphomedirection1) + str(fliphomedirection2)

        limitcont0 = struct.unpack('B', confstruct[108])[0]
        limitcont1 = struct.unpack('B', confstruct[109])[0]
        limitcont2 = struct.unpack('B', confstruct[110])[0]
        limitcont3 = struct.unpack('B', confstruct[111])[0]
        confdata += "\nLimitContCondition:"+str(limitcont0)+str(limitcont1) + \
                    str(limitcont2) + str(limitcont3)

        maxsteps0 = struct.unpack('<I', confstruct[112:116])[0]
        maxsteps1 = struct.unpack('<I', confstruct[116:120])[0]
        maxsteps2 = struct.unpack('<I', confstruct[120:124])[0]
        confdata += "\nMaxSteps:"+str(maxsteps0)+str(maxsteps1)+str(maxsteps2)

        tablesize0 = struct.unpack('<I', confstruct[124:128])[0]
        tablesize1 = struct.unpack('<I', confstruct[128:132])[0]
        tablesize2 = struct.unpack('<I', confstruct[132:136])[0]
        confdata += "\nTableSize:"+str(tablesize0)+str(tablesize1)+str(tablesize2)

        return confdata

    def calculatewh(self):
        """
        Calculates dimensions, in steps, of the current cut simulating
        the movement of the laser based on the commands, must run readcheckpacket() or checkpacket() previously
        """
        cmds = self.decompressed[:]
        cmds = cmds[8:]     # trims subheader
        self.commands = cmds[:]

        # extreme values
        maxheight = 0.
        maxwidth = 0.

        # initial position
        inipos = [0, 0]
        for n in range(len(cmds)/4):
                # the direction and magnitude of movement is determined
                xmove = -struct.unpack('b', cmds[4*n+1])[0]
                ymove = struct.unpack('b', cmds[4*n+2])[0]
                direction = struct.unpack('b', cmds[4*n])[0]

                # if is a no move command, skip
                if ymove == 0. and xmove == 0.:
                    continue

                # the sign of movement is changed according to direction byte
                if direction & 0b01:
                    xmove = -xmove
                if direction & 0b10:
                    ymove = -ymove

                # final position is calculated
                finpos = [xmove+inipos[0], ymove+inipos[1]]

                # next initial position
                inipos = finpos

                # updates max values
                if maxheight < abs(finpos[1]):
                    maxheight = abs(finpos[1])
                if maxwidth < abs(finpos[0]):
                    maxwidth = abs(finpos[0])

        self.height = maxheight
        self.width = maxwidth

    def readcheckpacket(self, namebinfile, writedec=False, separate=False):
        """Reads the bytes in a binary file, given the name of the file, returns a list where
        the first entry is a boolean that specifies whether there is a corrupted job(False)
        or not(True) and the second is a string with details about the job
        :param namebinfile: string that specifies the name of the binary file to read and check
        :param writedec: boolean that determines whether the decompressed job is written(True) or not(False)
        :param separate: boolean that determines whether the decompressed subpackets are written on different(True) or
        on the same file(False)
        :return: a list with the first entry a boolean that indicates if the packet is correct(True)
        and the second entry a string with details about the packet
        :rtype: list
        """
        # reads the binary
        self.binaryjob = open(namebinfile, 'r').read()

        # and checks
        return self.checkpacket(self.binaryjob, namebinfile, writedec, separate)

    def checkpacket(self, job, filename="Undefined", writedec=False, separate=False):
        """
        Checks the header and subpackets, returns a list with the first entry being a boolean that indicates if
        there is no error (True) and a second entry a string with details about the packet and errors
        :param job: string, is bynary job to check
        :param filename: string, will be part of the name of the generated files (.log and decompressed)
        :param writedec: boolean that determines whether the decompressed job is written(True) or not(False)
        :param separate: boolean that determines whether the decompressed subpackets are written on different(True) or
        on the same file(False)
        :return: a list with the first entry a boolean that indicates if the packet is correct(True)
        and the second entry a string with details about the packet
        :rtype: list
        """
        self.binaryjob = job

        try:
            # header to check
            header = self.binaryjob[:1024]

            # verifies first byte of header
            if header[0] == '\x02':
                self.packetdata += "\nVectorial Cut"
            else:
                self.packetdata += "\nWarning: It's not a vectorial cut"
                self.ok = False

            # stores the locations of bad bytes
            positions = []

            # check the zero bytes
            for n in range(1, 512) + range(513, 536) + range(669, 1024):
                if header[n] != '\x00':
                    positions.append(n)
                    self.ok = False

            # if errors then added to the log variable
            if not self.ok:
                self.packetdata += "\nAnormal byte(s) on: " + str(positions)

            # check bytes on 512 and 668
            if header[512] != '\x05':
                self.ok = False
                self.packetdata += "\nAnormal byte on 512 position, not 0x05"
            if header[668] != '\x17':
                self.ok = False
                self.packetdata += "\nAnormal byte on 668 position, not 0x17"

            # extracts the dimensions and configuration power from header
            self.widthheader = struct.unpack('<H', header[612: 614])[0]
            self.heightheader = struct.unpack('<H', header[616:618])[0]
            self.powerheader = struct.unpack('B', header[635])[0]

            # decompress the subpacket to check the subpackets
            self.decompressjob(filename, writedec, separate)

            # calculates the width and height
            self.countcommands()
            self.calculatewh()
        except IndexError:
            raise BadStellarisPacket()

        if not self.ok:
            raise BadStellarisPacket()

        return [self.ok, self.details(filename)]

    def decompressjob(self, fileout, writedecompressed=True, separate=False):
        """
        Decompresses the job, and may write it to file(s)
        :param fileout: string, will be part of the name of the written files(.log and decompressed)
        :param writedecompressed: boolean that determines whether the decompressed job is written(True) or not(False)
        :param separate: boolean that determines whether the decompressed subpackets are written on different(True) or
        on the same file(False)
        """
        self.decompressed = ""

        # obtains number of subpackets
        self.cantsubpack = struct.unpack('B', self.binaryjob[1024])[0]

        # checks the next three bytes
        if self.binaryjob[1025:1028] != '\x00\x00\x00':
            self.ok = False
            self.packetdata += "\nAnormal bytes on 1025 and 1027 positions" + str([self.binaryjob[1025:1028]])

        # controls the initial point where begins decompression
        offset = 1028

        # begins decompression of subpackets
        for sp in range(self.cantsubpack):
            # checks the separation bytes
            if self.binaryjob[offset + 2:offset + 4] != '\x00\x00':
                self.ok = False
                self.packetdata += "\nAnormal bytes on positions " + str([offset + 2, offset + 3])

            # determines the size of the subpacket
            sizesub = struct.unpack('H', self.binaryjob[offset:offset + 2])
            sizesub = sizesub[0]

            # obtains next subpacket
            sjob = self.binaryjob[offset + 4:]

            # the header for decompression is added
            sjob = "\x78\x9c" + sjob
            try:
                self.decompressed += (zlib.decompress(sjob))
            except zlib.error:
                writedecompressed = False
                self.ok = False
                self.packetdata += "\nError on decompressing the subpacket: " + str(sp)
                break

            # the offset is incremented to begin with next subpacket
            offset += sizesub + 4

        # in case of generatePack=True the decompressed bytes are written
        # to one file (separate=False) or to several files (separate=True)
        # one for each subpacket
        if writedecompressed:
            if separate:
                n = 1
                for p in self.decompressed:
                    output = open(fileout[:-4] + "_decompressed_pt" + str(n) + ".bin", "w")
                    output.write(p)
                    output.close()
                    n += 1
            else:
                output = open(fileout[:-4] + "_decompressed.bin", "w")
                output.write(arr2str(self.decompressed))
                output.close()

    def details(self, filename):
        """
        Returns important info about the checked job, and errors
        if any
        :param filename: filename of the checked packet
        :return: string with info about the packet and errors if any
        :rtype: str
        """
        det = "Job: " + filename
        det += "\nPower (according to header): " + str(self.powerheader)
        det += "\nWidth (according to header): " + str(self.widthheader)
        det += "\nHeight (according to header): " + str(self.heightheader)
        det += "\nWidth (according to decompressed): " + str(self.width)
        det += "\nHeight (according to decompressed): " + str(self.height)
        det += "\nNumber of supackets: " + str(self.cantsubpack)
        det += "\nNumber of commands: " + str(self.totalcmds)
        det += "\n\n" + str(self.datacmds)
        det += self.packetdata

        # in case of no errors OK is appended
        if self.ok:
            det += "\nOK\n"
        else:
            det += "\nERROR\n"

        return det

    def getmaxpower(self):
        """
        Finds the power of cutting in the checked packet, must run checkpacket() or readcheckpacket() previusly
        :return: power as a percentage
        :rtype: float
        """
        maxpower = 0
        for cmd in self.foundcmds.keys():
            power = struct.unpack('B', cmd[3])[0]
            if power > maxpower:
                maxpower = power
        return round(maxpower*100./255., 2)

    def getcuttime(self):
        """
        Calculates the estimated time of the cut, given that machine card executes 2000 commands per second
        :return: string in the format minutes:seconds
        """
        return time.strftime('%M:%S', time.gmtime(self.totalcmds*(1./2000.)))


class BadStellarisPacket(Exception):
    pass
