import svg.path
import scipy
import popplerqt4
import zlib
import math
import vectorcut
import cutterpath
from matplotlib.pyplot import *

################################################################
# Functions

def getangle(v1, v2):
    dotp = v1[0]*v2[0]+v1[1]*v2[1]
    magv1 = math.sqrt(v1[0]**2+v1[1]**2)
    magv2 = math.sqrt(v2[0]**2+v2[1]**2)
    return math.acos(dotp/(magv1*magv2))

def extract(filepath):
    """
        Reads a pdf file and gets the deflated stream
        :param filepath: the path to the pdf file
        :return: deflated stream, str type
        """

    # The extractor only works in documents with one page
    doc = popplerqt4.Poppler.Document.load(filepath)
    num_pages = doc.numPages()
    if num_pages > 1:
        raise WrongNumberOfPages

    pdffileobj = open(filepath, 'rb')

    # This Flag determines if the stream must be decompressed
    deflate = False

    # where will be the extracted stream from the pdf
    binstream = ""
    # Extract the binary stream from the pdf
    for line in pdffileobj:
        if line == '   /Filter /FlateDecode\n':
            deflate = True
        if line == 'stream\n':
            line = pdffileobj.next()
            while line != 'endstream\n':
                binstream = binstream + line
                line = pdffileobj.next()
    pdffileobj.close()

    # Deflate binary stream if necessary
    if deflate:
        deflatedstream = zlib.decompress(binstream)
    else:
        deflatedstream = binstream

    return deflatedstream


def _is_float_(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


################################################################

# Related to the generation of an intermediate representation of
# the path to do
class PathGenerator:

    def __init__(self, pdfpath, speed, power):
        """
        Initializes relevant variables
        :param pdfpath: path of the pdf file to read
        :param power: power of lines to cut
        """
        self.cuttingpath = cutterpath.CutterPath()
        self.deflatedstream = extract(pdfpath)
        self.currenpos = [0., 0.]
        self.startingpos = [0., 0.]
        self.stack = []
        self.speed = speed
        self.power = power
        self.laserstepsxy = [0, 0]
        self.movements = []


    # generates the path
    def generate(self):
        """
        Generate the full path from the deflated stream
        :return: CutterPath
        """
        pdfoperand = self.deflatedstream.split('\n')

        # process the objects in the deflated stream
        for operand in pdfoperand:
            for n in operand.split():
                self.operate(n)
        return self.cuttingpath

    def operate(self, op):
        """
        Receives an operator and chooses the apropiate method to manipulate the corresponding parameters
        :param op: operator from the deflated stream, type str
        """
        if _is_float_(op):
            self.stack.append(op)
        else:
            if op == 're':
                self._square_()

            # Moveto
            elif op == 'm':
                self._moveto_()

            # Lineto
            elif op == 'l':
                self._lineto_()

            # Curveto
            elif op == 'c':
                self._curveto_()

            # Close Current Path
            elif op == 'h':
                self._close_path_()
                # print "close path"

            else:
                pass
                # print 'Do nothing'

    def _moveto_(self):
        """
        This method takes 2 operands in the stack and form a point, then its generates
        vector using the Starting Position of the path as starting point and the produced point
        as ending point with a laser power of 0%.
        :return: Adds The generated Vector to the Current Path.
        """

        # pops the ending position, stores as inches
        ymovement = float(self.stack.pop()) / 72.
        xmovement = float(self.stack.pop()) / 72.

        # Move from the current position to [X,Y]
        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, [xmovement, ymovement], 100, 0.0))

        # updates position of current point, given the movement appended to the path
        self.updatepos([xmovement, ymovement])

        # Moveto always starts a New Path
        self.startingpos = self.currenpos[:]

    def _lineto_(self):
        """
        This method takes 2 operands in the stack and form a point using the first operand as the
        Y coordinate and the second as the X coordinate, then its generates vector using the CurrentPosition
        as starting point and the produced point as ending point.
        :return: Adds The generated Vector to the Path
        """

        # pops the ending position, stores as inches
        ymovement = float(self.stack.pop()) / 72.
        xmovement = float(self.stack.pop()) / 72.

        # add cut line to path
        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, [xmovement, ymovement], self.speed, self.power))

        # update position of current point, given the movement appended to the path
        self.updatepos([xmovement, ymovement])

    def _close_path_(self):
        """
        This Method Generates a Vector from the current position to the starting position of the path,
        closing the current path.
        :return: Adds The generated Vector to the Path and starts a new sub path.
        """

        # Closing the current sub path
        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, self.startingpos, self.speed, self.power))

        # updates current position
        self.updatepos(self.startingpos)

    def _square_(self):
        """
        Adds the necessary movements to the path to achieve a rectangle
        """

        # pops square parameters, as inches
        square_height = float(self.stack.pop()) / 72.
        square_width = float(self.stack.pop()) / 72.
        square_y0 = float(self.stack.pop()) / 72.
        square_x0 = float(self.stack.pop()) / 72.

        # defines points from square parameters
        point1 = [square_x0, square_y0]
        point2 = [square_x0 + square_width, square_y0]
        point3 = [square_x0 + square_width, square_y0 + square_height]
        point4 = [square_x0, square_y0 + square_height]

        # generates vector cuts from the defined points
        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, point1, self.speed))
        self.updatepos(point1)

        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, point2, self.speed, self.power))
        self.updatepos(point2)

        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, point3, self.speed, self.power))
        self.updatepos(point3)

        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, point4, self.speed, self.power))
        self.updatepos(point4)

        self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, point1, self.speed, self.power))
        self.updatepos(point1)

    def _curveto_(self):
        """
        Adds the neccesary vector cuts to make a curve
        """

        # Starting Point
        y0 = self.currenpos[1]
        x0 = self.currenpos[0]

        # Ending Point
        y3 = float(self.stack.pop()) / 72.
        x3 = float(self.stack.pop()) / 72.

        # Bezier Control Points
        y2 = float(self.stack.pop()) / 72.
        x2 = float(self.stack.pop()) / 72.
        y1 = float(self.stack.pop()) / 72.
        x1 = float(self.stack.pop()) / 72.


        # Setup the parametrizationc
        distance = (svg.path.CubicBezier(x0 + y0 * 1j, x1 + y1 * 1j, x2 + y2 * 1j, x3 + y3 * 1j).length(error=1e-5))
        vel = self.speed
        power = self.power

        if self.speed > 50.:
            branchsangle = getangle([x1-x0, y1-y0], [x3-x2, y3-y2])*180/3.1415
            if branchsangle <=90:
                fc = 50./self.speed # attenuation factor of speed
                vel = 50.   # best speed to close curves
                power = self.power * fc     # attenuates power given the attenuation factor

        # if self.speed > 80.:
        #     if distance*1000. <= 50:
        #         fc = 50./self.speed
        #         vel = 50.
        #         power = self.power * fc

        dotsperinc = 350  # parameter defined by trial and error
        number_of_points = int(distance * dotsperinc)  # number of points per curve
        t = scipy.linspace(0, 1, number_of_points)

        # Use the Bezier formula
        bx = (1 - t) ** 3 * x0 + 3 * (1 - t) ** 2 * t * x1 + 3 * (1 - t) * t ** 2 * x2 + t ** 3 * x3
        by = (1 - t) ** 3 * y0 + 3 * (1 - t) ** 2 * t * y1 + 3 * (1 - t) * t ** 2 * y2 + t ** 3 * y3

        bpc = 0  # Bezier Point Counter

        # Bezier Vector Generator
        for point in t:
            xpos = bx[bpc]
            ypos = by[bpc]
            nextpos = [xpos, ypos]
            self.cuttingpath.append(vectorcut.VectorCut(self.currenpos, nextpos, vel, power))
            self.updatepos(nextpos)
            bpc += 1

    def updatepos(self, endpt):
        """
        Predicts the laser position given a movement, can be said it acts as a sort of simulator
        :param endpt: ending point, given as a list of floats
        """
        self.currenpos = [round(endpt[0] - self.currenpos[0], 3)*1000 + self.currenpos[0]*1000,
                          round(endpt[1] - self.currenpos[1], 3)*1000 + self.currenpos[1]*1000]
        self.currenpos = [self.currenpos[0]/1000., self.currenpos[1]/1000.]

    def lineinterp(self, finish, power):
        xf=finish[0]
        yf=finish[1]

        #current position is set to the starting position
        x=self.laserstepsxy[0]
        y=self.laserstepsxy[1]

        #X Axis direction, if the value of DX is positive, the laser move to
        #the rigth if the value is negative the laser move to the left.

        dx=xf-x
        if dx >= 0:
           stepx=1
        else:
           stepx=-1

        #Y Axis direction, if the value of DY is positive, the laser move up
        #if the value is negative the laser down.

        dy=yf-y

        if dy >= 0:
           stepy=1
        else:
           stepy=-1

        #Current funtion symbol, if FXY is greater than or equal to zero, the
        #laser must move in the X axis, else the laser move in the Y axis
        fxy = math.fabs(dx) - math.fabs(dy)

        #Main loop:
        while not ((x==xf) and (y==yf)):

            if fxy >= 0:
                x = x + stepx
                fxy = fxy - math.fabs(dy)
                if fxy < 0:
                    y = y + stepy
                    fxy = fxy + math.fabs(dx)
                    self.movements += [[stepx, stepy], power]
                else:
                    self.movements += [[stepx, 0], power]
            else:
                y = y + stepy
                fxy = fxy + math.fabs(dx)
                if fxy >=0:
                    x = x + stepx
                    fxy = fxy - math.fabs(dy)
                    self.movements += [[stepx, stepy], power]
                else:
                    self.movements += [[0, stepy], power]
            #print('Current  point =  [{},{}]').format(x,y)
        return 0

    def delay_loop(speed):
        feed_rate = 1.0-(speed/100.0)
        counter = 0.0

        for i in range(0, 20):
            print i+1

            if counter >= 1.0:
                print "Wait\n"
                counter = 0.0

            else:
                print "command \n"

            counter += feed_rate


    #def cosineinterp(self):

#################################################################
# Custom exceptions

class Error(Exception):
    """Base class for other user defined exceptions"""
    pass


class WrongNumberOfPages(Error):
    """Raised when a multiple page PDF is selected to be parsed"""
    pass
